/************************************************************************
FICHERO estruc.h: Este fichero de cabecera contiene las declaraciones de
las estructuras que se emplearan en la aplicacion.
Son de uso comun en todos los ficheros de la aplicacion.
************************************************************************/

/* Estructura de datos asociada a un termino, ecuacion o regla.

Ejemplo de termino: f(g(a,X))

			       f 
			       |   
			       g    
			       |    
			       a - X

Ejemplo de ecuacion: f(g(a,X))=g(a,f(X))

			       f - g
			       |   |
			       g   a - f
			       |       |
			       a - X   X


Ejemplo de regla: f(g(a,X))=g(a,f(X)) <= X=f(g(Y,b)), g(a,f(b))=Y

			       f - g - X - f - g - Y
			       |   |       |   |    
			       g   a - f   g   a - f
			       |       |   |       |
			       a - X   X   Y - b   b 		*/
#include "time.h"

/* incluye capos para determinados criterios*/

PUBLIC struct arbol {char oc[10];float tam;int nejp;int nejn; struct arbol *herm, *hijo;}; 


/* Estructura de datos asociada a una teoria (conjunto de reglas) */

PUBLIC struct lista {struct arbol *ecu; struct lista *sig;};

/* Estructura de datos asociada a unas ocurrencias (conjunto de posiciones) */

PUBLIC struct ocus {struct oc_pos *posicion; struct ocus *sig;};
 
/* Estructura de datos asociada a una posicin */

PUBLIC struct oc_pos {int valor; struct oc_pos *sig;};

/* Estructura de datos asociada a unas ocurrencias (conjunto de posiciones) repetidas*/

PUBLIC struct ocus_rep {struct arbol *ocu; struct ocus *posis;struct ocus_rep *sig;};

/* Estructura de datos asociada a un programa (conjunto de reglas) */

PUBLIC struct prog {struct lista *prg; int nejp; int nejn; struct cobert *cob;
					float tam;double opt;int bkg;};

PUBLIC struct cobert {unsigned int mapa; struct cobert *sig;};
/* Estructura de datos asociada a un lista de programas (conjunto de programas) */ 

PUBLIC struct lprog {struct prog *pro; struct lprog *sig ;};

/* Estructura de datos asociada al conjunto de todos los 
nombres y aridades de las funciones que aparecen en la teoria :
Signatura. */

PUBLIC struct funcs {char nombre[10]; unsigned int aridad ; struct funcs *sig;};


/* Estructura de datos asociada a una substitucion (conjunto de
pares <variable, termino> */

PUBLIC struct subs {char var[10]; struct arbol *term; struct subs *sig;};

/* estrcutura para normalizar ecuaciones*/
struct list_var {char vell[10];char nou[10];struct list_var *sig;};

/* Estructura de datos asociada a un estado de Basic Narrowing
Calculus (conjunto de pares <ecuacion, sustitucion>, junto con
la ecuacion plana original de la que se ha derivado  */

PUBLIC struct stat { struct lista *ecus, *raiz; struct subs *alfa; struct stat *sig;};


/* Estructura de datos asociada a una lista de unificadores. Sera
utilizada en el calculo de la semantica de punto fijo  */

PUBLIC struct lisub { struct subs *alfa; struct lisub *sig;};


/* La siguiente estructura se usar en la realizacin de 
estadsticas sobre la eficiencia de los dos mtodos implementados
para la resolucin de objetivos, y los dos tipos de semnticas. */

PUBLIC  struct estadis {
char metodo;  /* Clculo efectuado por el Mtodo CLASICO (C) o SEMANTICO (S) */
unsigned instan;  /* Nmero de instanciaciones de una expresin por una sustitucin */
unsigned compos;  /* Nmero de composiciones de sustituciones. */
unsigned unifict;  /* Nmero total de operaciones de unificacin. */
unsigned unifice;  /* Nmero de operaciones de unificacin con exito . */
unsigned BN_RU1_TECAt; /* Nmero de operaciones BN1(), RU1() o TECA() totales. */
unsigned BN_RU1_TECAe; /* Nmero de operaciones BN1(), RU1() o TECA() con xito. */
unsigned BN_RU2t; /* Nmero de operaciones BN2() o RU2() totales. */
unsigned BN_RU2e; /* Nmero de operaciones BN2() o RU2() con xito. */
time_t time; /* Tiempo de ejecucin, en pulsos de reloj */
};


/* Esta estructura se emplea para referenciar mediante punteros los argumentos 
no variables de un termino funcion no plano */

struct tocur { 
	struct arbol *term, /* puntero a argumento no var. */
	*ant; /* puntero a arg. anterior a uno no var. */ 
	struct tocur *sig; 
	};
