/**************************************************************************
******************* RESOLUCION DE OBJETIVOS. ****************************
***************************************************************************
* FICHERO ob.c: Se implementa la resolucion de objetivos mediante dos metodos.
* El Metodo Clasico, empleando unificacion semantica por narrowing, y el
* Metodo Semntico empleando unificacion sintactica contra la semantica
* operacional O(E), o contra la semantica de punto fijo, F(E), ya que ambas
* coinciden. Los unicos puntos de entrada de este fichero son:
*
*	resob_cla(); Resolucion de objetivos mediante el Metodo Clasico.
*
*	resob_sem(); Resolucion de objetivos mediante el Metodo Semantico.
*
*******************************************************************************/
#include <stdio.h>
#include <string.h>
#include <ctype.h>
#include <stdlib.h>

#include "const.h"
#include "estruc.h"
#include "glovar.h"

/* Prototipos de funciones externas, definidas en Procesos auxiliares.
No todas estas funciones son empleadas en "ob.c". */ 
#include "proaux.h"

/* Prototipos de funciones externas, definidas en Fichero de Unificacion
Sintactica. No todas estas funciones son empleadas en "ob.c". */
#include "unific.h"

/* Prototipos de funciones externas, definidas en Fichero de Unificacion
Semantica. No todas estas funciones son empleadas en "ob.c". */
#include "narrow.h"


/******   RESOLUCION DE OBJETIVOS MEDIANTE EL METODO CLASICO  ********/

/* Toma como argumento la lista de estados del objetivo y genera el estado de 
narrowing inicial, que devuelve como resultado. El campo raiz no se necesita 
para nada, ya que el objetivo a tratar es unico. */

PUBLIC struct stat *genstat(struct lista *obj)
{
	struct stat *estaux=NULL;
	if (obj!=NULL) {
	  if ((estaux= (struct stat *) malloc(sizeof(struct stat))) == NULL)
					memerror();
	  estaux->ecus=copilis(obj); /* lista ecs. del objetivo */
	  estaux->raiz=NULL; /* No se necesita */
	  estaux->alfa=NULL; /* Substitucion indentidad, E */
	  estaux->sig=NULL;
	}
	return estaux;
}


/* Visualizacion de objetivos resueltos. Por la forma en que calcula las 
respuestas computadas el metodo de Basic Narrowing, basta con aplicar al 
objetivo inicial cada una de las sustituciones que contienen los estados 
de exito. No altera ninguna de las estructuras con las que trabaja. */

PUBLIC void verres(struct stat *exitos)
{
        int i; /* para estetica */
	struct arbol *e;
	struct lista *l; /* recorre la lista de ecuaciones del objetivo */
	struct stat *s; /* recorre la lista de estados de exito */
	if (exitos)
	{
	   fprintf(out, "\nRESULTADOS: INSTANCIACION DEL OBJETIVO.");
	   for (s=exitos, i=0; s; s=s->sig, i++) {
		/* visualiza la substitucion a aplicar */
		fprintf(out, "\n\nSUSTITUCION theta(%i) = ", i);
		verunif(s->alfa); fprintf(out,"\n");
		/* visualiza la instancia del objetivo */
		for (l=ejempini; l; l=l->sig) {
			e=copiecu(l->ecu);
			aplicar(&e, s->alfa);
			verecu(e);
			if (l != ejempfin) fprintf(out," , ");
			else fprintf(out,"\n");
			borraecu(e);
		}
	   }
	}
	else fprintf(out,"\nEL OBJETIVO ES IRRESOLUBLE.");
}


/******************************************************************************
Resolucion de un objetivo mediante el metodo clasico. Los pasos a seguir son
bien sencillos:

A)	Se genera un estado de narrowing inicial.
B)	Se aplica basic narrowing a dicho estado, para obtener las respuestas 
	computadas.
C)	Se visualizan las instancias del objetivo de acuerdo a las respuestas. 
*******************************************************************************/

PUBLIC void resob_cla()
{
	struct stat *S, /* puntero al estado de narrowing inicial */
		    *E; /* puntero a la lista de estados de exito.
			   Contienen las repuestas computadas. */
	start_st('C');
	S = genstat(ejempini);
	E = narrower(S,listaini,funescini,0); borrastat(S);
	verres(E); borrastat(E);
	stop_st();
}


