/*****************************************************************
***************** UNIFICACION SINTACTICA *************************
******************************************************************
FICHERO unific.c: Este modulo contiene los procedimientos que
permiten comprobar si dos expresiones son unificables. De serlo
se obtiene una sustitucion de maxima generalidad. Sus unicos puntos
de entrada son las funciones:

	aplicar():	Aplica una sustitucion a una expresion.

	componer():	Devuelve la composicion de dos sustituciones,
			que quedan inalteradas.

	unificador():	Calcula el unificador de maxima generalidad de
			dos expresiones.

************************************************************************/

#include <stdio.h>
#include <string.h>
#include <ctype.h>
#include <stdlib.h>

#include "const.h"
#include "estruc.h"
#include "glovar.h"

/* Prototipos de funciones externas, definidas en Procesos auxiliares.
No todas estas funciones son empleadas en "unific.c". */ 
#include "proaux.h"


/* Comprobar si la ocurrencia var aparece como subtermino en term */

PRIVATE int pertenece(char *var, struct arbol *term)
{
 if (term==NULL) return 0;
 else
   if (!strcmp(var,term->oc)) return 1;
   else return (pertenece(var, term->hijo) || pertenece(var,term->herm));
}


/* Aplicacion de una sustitucion de variable a una ecuacion  */

PRIVATE void paso(struct arbol **ecu, struct subs *alfa)
{
 struct arbol *termaux;
 if ((*ecu)!=NULL)
   { if(!strcmp((*ecu)->oc,alfa->var))
       {
	termaux=(*ecu)->herm;
	strcpy((*ecu)->oc,alfa->term->oc);
	(*ecu)->hijo=copiecu(alfa->term->hijo);
	(*ecu)->herm=termaux;
       }
     else  paso(&((*ecu)->hijo), alfa);
     paso(&((*ecu)->herm), alfa);
   }
}


/* Aplicacion de una sustitucion completa
 (varias variables) a una ecuacion. Cada vez
 que se instancia una variable se incrementa
 el contador correspondiente de estadisticas.  */

PUBLIC void aplicar(struct arbol **ecu, struct subs *alfa)
{
 for(;alfa!=NULL; paso(ecu,alfa), alfa=alfa->sig, st.instan++);
}


/*  Devuelve la composicion de las sustituciones
alfa y beta (quedando estas inalteradas). Cada vez
que se invoca al procedimiento se incrementa el
contador de estadisticas correspondiente. */

PUBLIC struct subs *componer  (struct subs *alfa, struct subs *beta)
{
 struct subs *alfaux, *subaux, *comp=NULL;
 struct arbol *termaux;
 st.compos++;
 for (alfaux=alfa;alfaux;alfaux=alfaux->sig)
    {
      termaux=copiterm(alfaux->term);
      aplicar(&termaux,beta);
      if (strcmp(alfaux->var,termaux->oc))
	 {
	    if ((subaux=(struct subs *)malloc(sizeof(struct subs)))==NULL)
				memerror();
	    strcpy(subaux->var,alfaux->var);
	    subaux->term=termaux;
	    subaux->sig=comp;
	    comp=subaux;
	 }
    }
 for ( ; beta; beta=beta->sig)
    {
      for ( alfaux=alfa;alfaux;alfaux=alfaux->sig)
	  if (!strcmp(alfaux->var,beta->var)) break;
      if (alfaux==NULL)
	 {
	    if ((subaux=(struct subs *)malloc(sizeof(struct subs)))==NULL)
				memerror();
	    strcpy(subaux->var,beta->var);
	    subaux->term=copiterm(beta->term);
	    subaux->sig=comp;
	    comp=subaux;
	 }
    }
 return comp;
}


/* Intenta unificar los subterminos a y b, que pertenecen
a los terminos raiza y raizb, quedando todos ellos modificados.
Si no existe el unificador de maxima generalidad de los terminos
apuntados por raiza y raizb, se devuelve un 0 y el parametro (*mgu)=NULL.
En otro caso se devuelve 1 y (*mgu) apunta al unificador de 
maxima generalidad. */

PRIVATE int unificar (struct arbol *raiza, struct arbol *raizb,
	      struct arbol *a, struct arbol *b, struct subs **mgu)
{

 struct subs *subaux;

 if ((a==NULL) && (b==NULL)) {	return 1; }
 else
  if ((a!=NULL) && (b!=NULL))
    if (!strcmp(a->oc,b->oc)) /* son iguales los primeros simbolos de a y b */
      if (unificar(raiza,raizb,a->hijo,b->hijo,mgu))
	  return unificar(raiza,raizb,a->herm,b->herm,mgu);
      else return 0;
    else
      if ((isupper(a->oc[0]))||(a->oc[0]=='_')) /* a es una variable */
       if ((isupper(b->oc[0]))||(b->oc[0]=='_'))  /* b es una variable */
	{
	  if ((subaux=(struct subs *)malloc(sizeof(struct subs)))==NULL)
				memerror();
	  strcpy(subaux->var,a->oc);
	  subaux->term=copiterm(b);
	  subaux->sig=(*mgu);
	  (*mgu)=subaux;
	  paso(&raiza,*mgu); paso(&raizb,*mgu);
	  return unificar(raiza,raizb,a->herm, b->herm, mgu);
	}
       else  /* a es variable pero b es una funcin */
	if (!pertenece(a->oc,b->hijo))
	  {
	    if ((subaux=(struct subs *)malloc(sizeof(struct subs)))==NULL)
				memerror();
	    strcpy(subaux->var,a->oc);
	    subaux->term=copiterm(b);
	    subaux->sig=(*mgu);
	    (*mgu)=subaux;
	    paso(&raiza,*mgu); paso(&raizb,*mgu);
	    return unificar(raiza,raizb,a->herm, b->herm, mgu);
	  }
	else return 0; /* la variable a esta contenida en b */
      else /* a no es una variable */
      if ((isupper(b->oc[0]))||(b->oc[0]=='_'))  /* b es una variable */
			      /* a es funcion y b es variable */
	if (!pertenece(b->oc,a->hijo))
	  {
	    if ((subaux=(struct subs *)malloc(sizeof(struct subs)))==NULL)
				memerror();
	    strcpy(subaux->var,b->oc);
	    subaux->term=copiterm(a);
	    subaux->sig=(*mgu);
	    (*mgu)=subaux;
	    paso(&raiza,*mgu); paso(&raizb,*mgu);
	    return unificar(raiza,raizb,a->herm, b->herm, mgu);
	  }
	 else return 0; /* la variable b esta contenida en a */
      else return 0; /* tanto a como b son funciones distintas */
  else return 0; /* a o b (uno solo de ellos) es vacio */
}


/* Calcula el unificador de maxima generalidad de las ecuaciones
a y b (que quedan inalteradas).Si no existe, se devuelve un 0 y el
parametro (*mgu)=NULL. En otro caso se devuelve 1 y (*mgu) apunta al
unificador de maxima generalidad. Cada vez que se invoca al procedimiento
se incrementa el contador estadistico de unificaciones totales, asi como
el de unificaciones con exito en los casos correspondientes. */

PUBLIC int unificador (struct arbol *a, struct arbol *b, struct subs **mgu)
{
  struct arbol *ecuauxa=NULL, *ecuauxb=NULL; 
  int i;
  st.unifict++;
  ecuauxa=copiecu(a); ecuauxb=copiecu(b);
  i=unificar(ecuauxa,ecuauxb,ecuauxa,ecuauxb,mgu);
  borraecu(ecuauxa); borraecu(ecuauxb);
  if (!i) st.unifice++;
  return i;
}



